use crate::{
    kw::{self, custom_bounds},
    Errors, HdlAttr, PairsIterExt,
};
use proc_macro2::{Span, TokenStream};
use quote::{quote_spanned, ToTokens};
use std::{collections::HashMap, fmt};
use syn::{
    parse::{Parse, ParseStream},
    parse_quote, parse_quote_spanned,
    punctuated::Punctuated,
    spanned::Spanned,
    token::{Brace, Paren},
    AngleBracketedGenericArguments, Attribute, ConstParam, Expr, Field, FieldMutability,
    FieldsNamed, GenericArgument, GenericParam, Generics, Ident, ImplGenerics, Path, PathArguments,
    PathSegment, PredicateType, Token, Turbofish, Type, TypeGenerics, TypeGroup, TypeParam,
    TypeParen, TypePath, TypeTuple, Visibility, WhereClause, WherePredicate,
};

crate::options! {
    #[options = ItemOptions]
    pub(crate) enum ItemOption {
        OutlineGenerated(outline_generated),
        ConnectInexact(connect_inexact),
        Target(target, Path),
        CustomBounds(custom_bounds),
        NoStatic(no_static),
    }
}

impl ItemOptions {
    pub(crate) fn validate(&mut self) -> syn::Result<()> {
        if let Self {
            custom_bounds: Some((custom_bounds,)),
            no_static: None,
            ..
        } = self
        {
            self.no_static = Some((kw::no_static(custom_bounds.span),));
        }
        Ok(())
    }
}

pub(crate) struct WrappedInConst<'a> {
    outer: &'a mut TokenStream,
    span: Span,
    inner: TokenStream,
}

impl<'a> WrappedInConst<'a> {
    pub(crate) fn new(outer: &'a mut TokenStream, span: Span) -> Self {
        Self {
            outer,
            span,
            inner: TokenStream::new(),
        }
    }
    pub(crate) fn inner(&mut self) -> &mut TokenStream {
        &mut self.inner
    }
}

impl Drop for WrappedInConst<'_> {
    fn drop(&mut self) {
        let inner = &self.inner;
        quote_spanned! {self.span=>
            const _: () = {
                #inner
            };
        }
        .to_tokens(self.outer);
    }
}

pub(crate) fn get_target(target: &Option<(kw::target, Paren, Path)>, item_ident: &Ident) -> Path {
    match target {
        Some((_, _, target)) => target.clone(),
        None => item_ident.clone().into(),
    }
}

pub(crate) fn common_derives(span: Span) -> Attribute {
    parse_quote_spanned! {span=>
        #[::fayalite::__std::prelude::v1::derive(
            ::fayalite::__std::fmt::Debug,
            ::fayalite::__std::cmp::Eq,
            ::fayalite::__std::cmp::PartialEq,
            ::fayalite::__std::hash::Hash,
            ::fayalite::__std::marker::Copy,
            ::fayalite::__std::clone::Clone,
        )]
    }
}

crate::options! {
    #[options = LifetimeParamOptions]
    enum LifetimeParamOption {}
}

crate::options! {
    #[options = TypeParamOptions]
    pub(crate) enum TypeParamOption {}
}

crate::options! {
    #[options = ConstParamOptions]
    pub(crate) enum ConstParamOption {}
}

crate::options! {
    #[options = TypeOptions]
    enum TypeOption {}
}

macro_rules! parse_failed {
    ($parser:ident, $tokens:expr, $message:expr) => {{
        $parser.errors().error($tokens, $message);
        return Err(ParseFailed);
    }};
}

#[derive(Copy, Clone, Debug)]
pub(crate) enum TypeDelimiter {
    Group(syn::token::Group),
    Paren(Paren),
}

impl TypeDelimiter {
    pub(crate) fn surround<F: FnOnce(&mut TokenStream)>(self, tokens: &mut TokenStream, f: F) {
        match self {
            TypeDelimiter::Group(v) => v.surround(tokens, f),
            TypeDelimiter::Paren(v) => v.surround(tokens, f),
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedTypeDelimited {
    pub(crate) delim: TypeDelimiter,
    pub(crate) elem: Box<ParsedType>,
}

impl ToTokens for ParsedTypeDelimited {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let Self { delim, elem } = self;
        delim.surround(tokens, |tokens| elem.to_tokens(tokens));
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedTypeTuple {
    pub(crate) paren_token: Paren,
    pub(crate) elems: Punctuated<ParsedType, Token![,]>,
}

impl ToTokens for ParsedTypeTuple {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let Self { paren_token, elems } = self;
        paren_token.surround(tokens, |tokens| {
            elems.to_tokens(tokens);
            if elems.len() == 1 && !elems.trailing_punct() {
                // trailing comma needed to distinguish from just parenthesis
                Token![,](paren_token.span.close()).to_tokens(tokens);
            }
        })
    }
}

impl ParseTypes<TypeTuple> for ParsedTypeTuple {
    fn parse_types(ty: &mut TypeTuple, parser: &mut TypesParser<'_>) -> Result<Self, ParseFailed> {
        Ok(Self {
            paren_token: ty.paren_token,
            elems: parser.parse(&mut ty.elems)?,
        })
    }
}

#[derive(Debug, Clone)]
pub(crate) enum ParsedGenericArgument {
    Type(ParsedType),
    Const(Expr),
}

impl ToTokens for ParsedGenericArgument {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        match self {
            Self::Type(ty) => ty.to_tokens(tokens),
            Self::Const(expr) => expr.to_tokens(tokens),
        }
    }
}

impl ParseTypes<GenericArgument> for ParsedGenericArgument {
    fn parse_types(
        arg: &mut GenericArgument,
        parser: &mut TypesParser<'_>,
    ) -> Result<Self, ParseFailed> {
        match arg {
            GenericArgument::Type(ty) => Ok(Self::Type(parser.parse(ty)?)),
            GenericArgument::Const(expr) => Ok(Self::Const(expr.clone())),
            _ => parse_failed!(parser, arg, "expected type or const generic argument"),
        }
    }
}

impl From<ParsedGenericArgument> for GenericArgument {
    fn from(value: ParsedGenericArgument) -> Self {
        match value {
            ParsedGenericArgument::Type(ty) => Self::Type(ty.into()),
            ParsedGenericArgument::Const(expr) => Self::Const(expr),
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedGenericArguments {
    pub(crate) colon2_token: Option<Token![::]>,
    pub(crate) lt_token: Token![<],
    pub(crate) args: Punctuated<ParsedGenericArgument, Token![,]>,
    pub(crate) gt_token: Token![>],
}

impl ToTokens for ParsedGenericArguments {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let Self {
            colon2_token,
            lt_token,
            args,
            gt_token,
        } = self;
        colon2_token.to_tokens(tokens);
        lt_token.to_tokens(tokens);
        args.to_tokens(tokens);
        gt_token.to_tokens(tokens);
    }
}

impl ParseTypes<AngleBracketedGenericArguments> for ParsedGenericArguments {
    fn parse_types(
        args: &mut AngleBracketedGenericArguments,
        parser: &mut TypesParser<'_>,
    ) -> Result<Self, ParseFailed> {
        let AngleBracketedGenericArguments {
            colon2_token,
            lt_token,
            ref mut args,
            gt_token,
        } = *args;
        Ok(Self {
            colon2_token,
            lt_token,
            args: parser.parse(args)?,
            gt_token,
        })
    }
}

impl From<ParsedGenericArguments> for AngleBracketedGenericArguments {
    fn from(value: ParsedGenericArguments) -> Self {
        let ParsedGenericArguments {
            colon2_token,
            lt_token,
            args,
            gt_token,
        } = value;
        AngleBracketedGenericArguments {
            colon2_token,
            lt_token,
            args: Punctuated::from_iter(args.into_pairs().map_pair_value(Into::into)),
            gt_token,
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedTypeNamed {
    pub(crate) leading_colon: Option<Token![::]>,
    pub(crate) segments: Punctuated<Ident, Token![::]>,
    pub(crate) args: Option<ParsedGenericArguments>,
}

impl ParsedTypeNamed {
    pub(crate) fn get_ident(&self) -> Option<&Ident> {
        let Self {
            leading_colon: None,
            segments,
            args: None,
        } = self
        else {
            return None;
        };
        if segments.len() == 1 && !segments.trailing_punct() {
            Some(&segments[0])
        } else {
            None
        }
    }
}

impl ToTokens for ParsedTypeNamed {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let Self {
            leading_colon,
            segments,
            args,
        } = self;
        leading_colon.to_tokens(tokens);
        segments.to_tokens(tokens);
        args.to_tokens(tokens);
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedTypeNamedParamType {
    pub(crate) ident: Ident,
    pub(crate) param_index: usize,
}

impl ToTokens for ParsedTypeNamedParamType {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let Self {
            ident,
            param_index: _,
        } = self;
        ident.to_tokens(tokens);
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedTypeNamedParamSizeType {
    pub(crate) ident: Ident,
    pub(crate) param_index: usize,
}

impl ToTokens for ParsedTypeNamedParamSizeType {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let Self {
            ident,
            param_index: _,
        } = self;
        ident.to_tokens(tokens);
    }
}

#[derive(Debug, Clone)]
pub(crate) enum ParsedTypeNamedParam {
    Type(ParsedTypeNamedParamType),
    SizeType(ParsedTypeNamedParamSizeType),
}

impl ParsedTypeNamedParam {
    pub(crate) fn into_ident(self) -> Ident {
        match self {
            Self::Type(v) => v.ident,
            Self::SizeType(v) => v.ident,
        }
    }
    pub(crate) fn ident(&self) -> &Ident {
        match self {
            Self::Type(v) => &v.ident,
            Self::SizeType(v) => &v.ident,
        }
    }
    pub(crate) fn param_index(&self) -> usize {
        match self {
            Self::Type(v) => v.param_index,
            Self::SizeType(v) => v.param_index,
        }
    }
}

impl ToTokens for ParsedTypeNamedParam {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        match self {
            Self::Type(v) => v.to_tokens(tokens),
            Self::SizeType(v) => v.to_tokens(tokens),
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) enum ParsedType {
    Delimited(ParsedTypeDelimited),
    Named(ParsedTypeNamed),
    NamedParam(ParsedTypeNamedParam),
    Tuple(ParsedTypeTuple),
}

impl From<ParsedType> for Type {
    fn from(value: ParsedType) -> Self {
        match value {
            ParsedType::Delimited(ParsedTypeDelimited {
                delim: TypeDelimiter::Group(group_token),
                elem,
            }) => Type::Group(TypeGroup {
                group_token,
                elem: Box::new(elem.into()),
            }),
            ParsedType::Delimited(ParsedTypeDelimited {
                delim: TypeDelimiter::Paren(paren_token),
                elem,
            }) => Type::Paren(TypeParen {
                paren_token,
                elem: Box::new(elem.into()),
            }),
            ParsedType::NamedParam(value) => Type::Path(TypePath {
                qself: None,
                path: value.into_ident().into(),
            }),
            ParsedType::Named(ParsedTypeNamed {
                leading_colon,
                segments,
                args,
            }) => Type::Path(TypePath {
                qself: None,
                path: Path {
                    leading_colon,
                    segments: {
                        let mut segments = Punctuated::<PathSegment, _>::from_iter(
                            segments.into_pairs().map_pair_value(Into::into),
                        );
                        if let Some(args) = args {
                            segments.last_mut().unwrap().arguments =
                                PathArguments::AngleBracketed(args.into());
                        }
                        segments
                    },
                },
            }),
            ParsedType::Tuple(ParsedTypeTuple { paren_token, elems }) => Type::Tuple(TypeTuple {
                paren_token,
                elems: Punctuated::from_iter(elems.into_pairs().map_pair_value(Into::into)),
            }),
        }
    }
}

impl From<MaybeParsed<ParsedType, Type>> for Type {
    fn from(value: MaybeParsed<ParsedType, Type>) -> Self {
        match value {
            MaybeParsed::Unrecognized(value) => value,
            MaybeParsed::Parsed(value) => value.into(),
        }
    }
}

impl From<Box<ParsedType>> for Type {
    fn from(value: Box<ParsedType>) -> Self {
        (*value).into()
    }
}

impl ToTokens for ParsedType {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        match self {
            ParsedType::Delimited(ty) => ty.to_tokens(tokens),
            ParsedType::NamedParam(ty) => ty.to_tokens(tokens),
            ParsedType::Named(ty) => ty.to_tokens(tokens),
            ParsedType::Tuple(ty) => ty.to_tokens(tokens),
        }
    }
}

impl ParsedType {
    pub(crate) fn unwrap_delimiters(mut self) -> Self {
        loop {
            match self {
                ParsedType::Delimited(ty) => self = *ty.elem,
                _ => return self,
            }
        }
    }
    pub(crate) fn unwrap_delimiters_ref(mut self: &Self) -> &Self {
        loop {
            match self {
                ParsedType::Delimited(ty) => self = &*ty.elem,
                _ => return self,
            }
        }
    }
    pub(crate) fn unwrap_delimiters_mut(mut self: &mut Self) -> &mut Self {
        loop {
            match self {
                ParsedType::Delimited(ty) => self = &mut *ty.elem,
                _ => return self,
            }
        }
    }
}

impl ParseTypes<Path> for ParsedType {
    fn parse_types(path: &mut Path, parser: &mut TypesParser<'_>) -> Result<Self, ParseFailed> {
        let Path {
            leading_colon,
            ref mut segments,
        } = *path;
        if segments.is_empty() {
            parse_failed!(parser, path, "path must not be empty");
        }
        let mut args = None;
        let segments = Punctuated::from_iter(segments.pairs_mut().map_pair_value_mut(|segment| {
            let PathSegment { ident, arguments } = segment;
            if let Some(_) = args {
                parser
                    .errors()
                    .error(&ident, "associated types/consts are not yet implemented");
            }
            args = match arguments {
                PathArguments::None => None,
                PathArguments::AngleBracketed(args) => parser.parse(args).ok(),
                PathArguments::Parenthesized(_) => {
                    parser
                        .errors()
                        .error(&segment, "function traits are not allowed");
                    args = None;
                    return segment.ident.clone();
                }
            };
            ident.clone()
        }));
        let named = ParsedTypeNamed {
            leading_colon,
            segments,
            args,
        };
        if let Some(ident) = named.get_ident() {
            if let Some(&param_index) = parser.generics.param_name_to_index_map.get(ident) {
                if parser
                    .cur_param_index
                    .is_some_and(|cur_param_index| param_index >= cur_param_index)
                {
                    parser.errors.error(
                        parser.generics.params[param_index].ident(),
                        "cannot use forward declared identifier in generic parameter defaults",
                    );
                }
                return Ok(Self::NamedParam(
                    match parser.generics.params[param_index] {
                        ParsedGenericParam::Type(_) => {
                            ParsedTypeNamedParam::Type(ParsedTypeNamedParamType {
                                ident: ident.clone(),
                                param_index,
                            })
                        }
                        ParsedGenericParam::SizeType(_) => {
                            ParsedTypeNamedParam::SizeType(ParsedTypeNamedParamSizeType {
                                ident: ident.clone(),
                                param_index,
                            })
                        }
                    },
                ));
            }
        }
        Ok(Self::Named(named))
    }
}

impl ParseTypes<TypePath> for ParsedType {
    fn parse_types(ty: &mut TypePath, parser: &mut TypesParser<'_>) -> Result<Self, ParseFailed> {
        let TypePath { qself, path } = ty;
        if let Some(qself) = qself {
            // TODO: implement
            parse_failed!(
                parser,
                ty,
                "associated types/consts are not yet implemented"
            );
        } else {
            parser.parse(path)
        }
    }
}

impl ParseTypes<Type> for ParsedType {
    fn parse_types(ty: &mut Type, parser: &mut TypesParser<'_>) -> Result<Self, ParseFailed> {
        Ok(match ty {
            Type::Array(_) => parse_failed!(parser, ty, "array types are not allowed here"),
            Type::BareFn(_) => parse_failed!(parser, ty, "fn() types are not allowed here"),
            Type::Group(TypeGroup { group_token, elem }) => Self::Delimited(ParsedTypeDelimited {
                delim: TypeDelimiter::Group(*group_token),
                elem: parser.parse(elem)?,
            }),
            Type::ImplTrait(_) => {
                parse_failed!(parser, ty, "impl Trait types are not allowed here")
            }
            Type::Infer(_) => parse_failed!(parser, ty, "inferred types are not allowed here"),
            Type::Macro(_) => parse_failed!(parser, ty, "macro types are not allowed here"),
            Type::Never(_) => parse_failed!(parser, ty, "the never type is not allowed here"),
            Type::Paren(TypeParen { paren_token, elem }) => Self::Delimited(ParsedTypeDelimited {
                delim: TypeDelimiter::Paren(*paren_token),
                elem: parser.parse(elem)?,
            }),
            Type::Path(ty) => parser.parse(ty)?,
            Type::Ptr(_) => parse_failed!(parser, ty, "pointer types are not allowed here"),
            Type::Reference(_) => parse_failed!(parser, ty, "reference types are not allowed here"),
            Type::Slice(_) => parse_failed!(parser, ty, "slice types are not allowed here"),
            Type::TraitObject(_) => {
                parse_failed!(parser, ty, "dyn Trait types are not allowed here")
            }
            Type::Tuple(ty) => Self::Tuple(parser.parse(ty)?),
            Type::Verbatim(_) => parse_failed!(parser, ty, "unknown type kind"),
            _ => parse_failed!(parser, ty, "unknown type kind"),
        })
    }
}

pub(crate) struct ParseFailed;

pub(crate) struct TypesParser<'a> {
    generics: &'a ParsedGenerics,
    /// used when parsing generic parameter defaults so
    /// earlier parameter defaults can't refer to later parameters
    cur_param_index: Option<usize>,
    errors: &'a mut Errors,
}

impl<'a> TypesParser<'a> {
    pub(crate) fn run_with_errors<T: ParseTypes<I>, I>(
        generics: &ParsedGenerics,
        input: &mut I,
        errors: &mut Errors,
    ) -> Result<T, ParseFailed> {
        TypesParser {
            generics,
            cur_param_index: None,
            errors,
        }
        .parse(input)
    }
    pub(crate) fn run<T: ParseTypes<I>, I>(
        generics: &ParsedGenerics,
        input: &mut I,
    ) -> syn::Result<T> {
        let mut errors = Errors::new();
        let retval = Self::run_with_errors(generics, input, &mut errors);
        errors.finish()?;
        retval.map_err(|ParseFailed {}| panic!("parse failed without registering any errors"))
    }
    pub(crate) fn maybe_run<T: ParseTypes<I>, I, G>(
        generics: MaybeParsed<&ParsedGenerics, G>,
        mut input: I,
        errors: &mut Errors,
    ) -> MaybeParsed<T, I> {
        let MaybeParsed::Parsed(generics) = generics else {
            return MaybeParsed::Unrecognized(input);
        };
        match Self::run_with_errors(generics, &mut input, errors) {
            Ok(v) => MaybeParsed::Parsed(v),
            Err(ParseFailed {}) => MaybeParsed::Unrecognized(input),
        }
    }
    pub(crate) fn generics(&self) -> &'a ParsedGenerics {
        self.generics
    }
    pub(crate) fn errors(&mut self) -> &mut Errors {
        self.errors
    }
    pub(crate) fn parse<T: ParseTypes<I>, I>(&mut self, input: &mut I) -> Result<T, ParseFailed> {
        T::parse_types(input, self)
    }
}

pub(crate) trait ParseTypes<I>: Sized {
    fn parse_types(input: &mut I, parser: &mut TypesParser<'_>) -> Result<Self, ParseFailed>;
}

impl<T: ParseTypes<I>, I> ParseTypes<Box<I>> for Box<T> {
    fn parse_types(input: &mut Box<I>, parser: &mut TypesParser<'_>) -> Result<Self, ParseFailed> {
        Ok(Box::new(parser.parse(&mut **input)?))
    }
}

impl<T: ParseTypes<I>, I, P: Clone> ParseTypes<Punctuated<I, P>> for Punctuated<T, P> {
    fn parse_types(
        input: &mut Punctuated<I, P>,
        parser: &mut TypesParser<'_>,
    ) -> Result<Self, ParseFailed> {
        let retval = Punctuated::from_iter(
            input
                .pairs_mut()
                .filter_map_pair_value_mut(|input| parser.parse(input).ok()),
        );
        Ok(retval)
    }
}

#[derive(Debug, Clone)]
pub(crate) enum UnparsedGenericParam {
    Type {
        attrs: Vec<Attribute>,
        options: HdlAttr<TypeParamOptions>,
        ident: Ident,
        colon_token: Token![:],
        bounds: ParsedBounds,
    },
}

pub(crate) mod known_items {
    use proc_macro2::{Ident, Span, TokenStream};
    use quote::ToTokens;
    use syn::{
        parse::{Parse, ParseStream},
        Path, PathArguments, PathSegment, Token,
    };

    macro_rules! impl_known_item_body {
        ($known_item:ident) => {
            #[derive(Clone, Debug)]
            pub(crate) struct $known_item {
                pub(crate) path: Path,
            }

            #[allow(non_snake_case, dead_code)]
            pub(crate) fn $known_item(span: Span) -> $known_item {
                let segments = $known_item::PATH_SEGMENTS.iter()
                    .copied()
                    .map(|seg| PathSegment::from(Ident::new(seg, span)))
                    .collect();
                $known_item {
                    path: Path {
                        leading_colon: Some(Token![::](span)),
                        segments,
                    },
                }
            }

            impl ToTokens for $known_item {
                fn to_tokens(&self, tokens: &mut TokenStream) {
                    self.path.to_tokens(tokens);
                }
            }

            impl $known_item {
                pub(crate) fn parse_path(path: Path) -> Result<Self, Path> {
                    if path.is_ident(stringify!($known_item)) {
                        Ok(Self { path })
                    } else if path.segments.len() == Self::PATH_SEGMENTS.len()
                        && path
                            .segments
                            .iter()
                            .zip(Self::PATH_SEGMENTS)
                            .all(|(seg, expected)| {
                                matches!(seg.arguments, PathArguments::None)
                                && seg.ident == *expected
                            })
                    {
                        Ok(Self { path })
                    } else {
                        Err(path)
                    }
                }
            }

            impl Parse for $known_item {
                fn parse(input: ParseStream) -> syn::Result<Self> {
                    Self::parse_path(Path::parse_mod_style(input)?).map_err(|path| {
                        syn::Error::new_spanned(
                            path,
                            concat!("expected ", stringify!($known_item)),
                        )
                    })
                }
            }
        };
    }

    macro_rules! impl_known_item {
        ($([$(::$head:ident)*])? ::$next:ident $(::$tail:ident)+) => {
            impl_known_item!([$($(::$head)*)? ::$next] $(::$tail)+);
        };
        ([$(::$seg:ident)+] ::$known_item:ident) => {
            impl_known_item_body!($known_item);

            impl $known_item {
                pub(crate) const PATH_SEGMENTS: &'static [&'static str] = &[
                    $(stringify!($seg),)+
                    stringify!($known_item),
                ];
            }
        };
    }

    impl_known_item!(::fayalite::bundle::BundleType);
    impl_known_item!(::fayalite::enum_::EnumType);
    impl_known_item!(::fayalite::int::IntType);
    impl_known_item!(::fayalite::int::KnownSize);
    impl_known_item!(::fayalite::int::Size);
    impl_known_item!(::fayalite::ty::StaticType);
    impl_known_item!(::fayalite::ty::Type);
}

macro_rules! impl_bounds {
    (
        #[struct = $struct_type:ident]
        $vis:vis enum $enum_type:ident {
            $(
                $Variant:ident,
            )*
        }
    ) => {
        #[derive(Clone, Debug)]
        $vis enum $enum_type {
            $($Variant(known_items::$Variant),)*
        }

        $(impl From<known_items::$Variant> for $enum_type {
            fn from(v: known_items::$Variant) -> Self {
                Self::$Variant(v)
            }
        })*

        impl ToTokens for $enum_type {
            fn to_tokens(&self, tokens: &mut TokenStream) {
                match self {
                    $(Self::$Variant(v) => v.to_tokens(tokens),)*
                }
            }
        }

        impl $enum_type {
            $vis fn parse_path(path: Path) -> Result<Self, Path> {
                $(let path = match known_items::$Variant::parse_path(path) {
                    Ok(v) => return Ok(Self::$Variant(v)),
                    Err(path) => path,
                };)*
                Err(path)
            }
        }

        impl Parse for $enum_type {
            fn parse(input: ParseStream) -> syn::Result<Self> {
                Self::parse_path(Path::parse_mod_style(input)?).map_err(|path| {
                    syn::Error::new_spanned(
                        path,
                        format_args!("expected one of: {}", [$(stringify!($Variant)),*].join(", ")),
                    )
                })
            }
        }

        #[derive(Clone, Debug, Default)]
        #[allow(non_snake_case)]
        $vis struct $struct_type {
            $($vis $Variant: Option<known_items::$Variant>,)*
        }

        impl ToTokens for $struct_type {
            #[allow(unused_mut, unused_variables, unused_assignments)]
            fn to_tokens(&self, tokens: &mut TokenStream) {
                let mut separator = None;
                $(if let Some(v) = &self.$Variant {
                    separator.to_tokens(tokens);
                    separator = Some(<Token![+]>::default());
                    v.to_tokens(tokens);
                })*
            }
        }

        const _: () = {
            #[derive(Clone, Debug)]
            $vis struct Iter($vis $struct_type);

            impl IntoIterator for $struct_type {
                type Item = $enum_type;
                type IntoIter = Iter;

                fn into_iter(self) -> Self::IntoIter {
                    Iter(self)
                }
            }

            impl Iterator for Iter {
                type Item = $enum_type;


                fn next(&mut self) -> Option<Self::Item> {
                    $(
                        if let Some(value) = self.0.$Variant.take() {
                            return Some($enum_type::$Variant(value));
                        }
                    )*
                    None
                }

                #[allow(unused_mut, unused_variables)]
                fn fold<B, F: FnMut(B, Self::Item) -> B>(mut self, mut init: B, mut f: F) -> B {
                    $(
                        if let Some(value) = self.0.$Variant.take() {
                            init = f(init, $enum_type::$Variant(value));
                        }
                    )*
                    init
                }
            }
        };

        impl Extend<$enum_type> for $struct_type {
            fn extend<T: IntoIterator<Item = $enum_type>>(&mut self, iter: T) {
                iter.into_iter().for_each(|v| match v {
                    $($enum_type::$Variant(v) => {
                        self.$Variant = Some(v);
                    })*
                });
            }
        }

        impl FromIterator<$enum_type> for $struct_type {
            fn from_iter<T: IntoIterator<Item = $enum_type>>(iter: T) -> Self {
                let mut retval = Self::default();
                retval.extend(iter);
                retval
            }
        }

        impl Extend<$struct_type> for $struct_type {
            fn extend<T: IntoIterator<Item = $struct_type>>(&mut self, iter: T) {
                iter.into_iter().for_each(|v| {
                    $(if let Some(v) = v.$Variant {
                        self.$Variant = Some(v);
                    })*
                });
            }
        }

        impl FromIterator<$struct_type> for $struct_type {
            fn from_iter<T: IntoIterator<Item = $struct_type>>(iter: T) -> Self {
                let mut retval = Self::default();
                retval.extend(iter);
                retval
            }
        }

        impl Parse for $struct_type {
            fn parse(input: ParseStream) -> syn::Result<Self> {
                let mut retval = Self::default();
                while !input.is_empty() {
                    retval.extend([input.parse::<$enum_type>()?]);
                    if input.is_empty() {
                        break;
                    }
                    input.parse::<Token![+]>()?;
                }
                Ok(retval)
            }
        }
    };
}

impl_bounds! {
    #[struct = ParsedBounds]
    pub(crate) enum ParsedBound {
        BundleType,
        EnumType,
        IntType,
        KnownSize,
        Size,
        StaticType,
        Type,
    }
}

impl_bounds! {
    #[struct = ParsedTypeBounds]
    pub(crate) enum ParsedTypeBound {
        BundleType,
        EnumType,
        IntType,
        StaticType,
        Type,
    }
}

impl_bounds! {
    #[struct = ParsedSizeTypeBounds]
    pub(crate) enum ParsedSizeTypeBound {
        KnownSize,
        Size,
    }
}

#[derive(Clone, Debug)]
pub(crate) enum ParsedBoundsCategory {
    Type(ParsedTypeBounds),
    SizeType(ParsedSizeTypeBounds),
}

impl ParsedBounds {
    fn categorize(self, errors: &mut Errors, span: Span) -> ParsedBoundsCategory {
        let mut type_bounds = None;
        let mut size_type_bounds = None;
        self.into_iter().for_each(|bound| match bound.categorize() {
            ParsedBoundCategory::Type(bound) => {
                type_bounds
                    .get_or_insert_with(ParsedTypeBounds::default)
                    .extend([bound]);
            }
            ParsedBoundCategory::SizeType(bound) => {
                size_type_bounds
                    .get_or_insert_with(ParsedSizeTypeBounds::default)
                    .extend([bound]);
            }
        });
        match (type_bounds, size_type_bounds) {
            (None, None) => ParsedBoundsCategory::Type(ParsedTypeBounds {
                Type: Some(known_items::Type(span)),
                ..Default::default()
            }),
            (None, Some(bounds)) => ParsedBoundsCategory::SizeType(bounds),
            (Some(bounds), None) => ParsedBoundsCategory::Type(bounds),
            (Some(type_bounds), Some(size_type_bounds)) => {
                errors.error(
                    size_type_bounds
                        .Size
                        .unwrap_or_else(|| known_items::Size(span)),
                    "conflicting bounds: can't use `Size` bounds with `Type` bounds",
                );
                ParsedBoundsCategory::Type(type_bounds)
            }
        }
    }
}

#[derive(Clone, Debug)]
pub(crate) enum ParsedBoundCategory {
    Type(ParsedTypeBound),
    SizeType(ParsedSizeTypeBound),
}

impl ParsedBound {
    fn categorize(self) -> ParsedBoundCategory {
        match self {
            Self::BundleType(v) => ParsedBoundCategory::Type(ParsedTypeBound::BundleType(v)),
            Self::EnumType(v) => ParsedBoundCategory::Type(ParsedTypeBound::EnumType(v)),
            Self::IntType(v) => ParsedBoundCategory::Type(ParsedTypeBound::IntType(v)),
            Self::KnownSize(v) => ParsedBoundCategory::SizeType(ParsedSizeTypeBound::KnownSize(v)),
            Self::Size(v) => ParsedBoundCategory::SizeType(ParsedSizeTypeBound::Size(v)),
            Self::StaticType(v) => ParsedBoundCategory::Type(ParsedTypeBound::StaticType(v)),
            Self::Type(v) => ParsedBoundCategory::Type(ParsedTypeBound::Type(v)),
        }
    }
    fn implied_bounds(self) -> ParsedBounds {
        let span = self.span();
        match self {
            Self::BundleType(v) => ParsedBounds::from_iter([
                ParsedBound::from(v),
                ParsedBound::Type(known_items::Type(span)),
            ]),
            Self::EnumType(v) => ParsedBounds::from_iter([
                ParsedBound::from(v),
                ParsedBound::Type(known_items::Type(span)),
            ]),
            Self::IntType(v) => ParsedBounds::from_iter([
                ParsedBound::from(v),
                ParsedBound::Type(known_items::Type(span)),
            ]),
            Self::KnownSize(v) => ParsedBounds::from_iter([
                ParsedBound::from(v),
                ParsedBound::Size(known_items::Size(span)),
            ]),
            Self::Size(v) => ParsedBounds::from_iter([ParsedBound::from(v)]),
            Self::StaticType(v) => ParsedBounds::from_iter([
                ParsedBound::from(v),
                ParsedBound::Type(known_items::Type(span)),
            ]),
            Self::Type(v) => ParsedBounds::from_iter([ParsedBound::from(v)]),
        }
    }
}

impl ParsedBounds {
    fn add_implied_bounds(&mut self) {
        let orig_bounds = self.clone();
        self.extend(self.clone().into_iter().map(ParsedBound::implied_bounds));
        self.extend([orig_bounds]); // keep spans of explicitly provided bounds
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedTypeParam {
    pub(crate) attrs: Vec<Attribute>,
    pub(crate) options: HdlAttr<TypeParamOptions>,
    pub(crate) ident: Ident,
    pub(crate) colon_token: Token![:],
    pub(crate) bounds: ParsedTypeBounds,
    pub(crate) default: Option<(Token![=], ParsedType)>,
}

impl ToTokens for ParsedTypeParam {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let Self {
            attrs,
            options,
            ident,
            colon_token,
            bounds,
            default,
        } = self;
        let TypeParamOptions {} = options.body;
        for attr in attrs {
            attr.to_tokens(tokens);
        }
        ident.to_tokens(tokens);
        colon_token.to_tokens(tokens);
        bounds.to_tokens(tokens);
        if let Some((eq, ty)) = default {
            eq.to_tokens(tokens);
            ty.to_tokens(tokens);
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedSizeTypeParam {
    pub(crate) attrs: Vec<Attribute>,
    pub(crate) options: HdlAttr<TypeParamOptions>,
    pub(crate) ident: Ident,
    pub(crate) colon_token: Token![:],
    pub(crate) bounds: ParsedSizeTypeBounds,
    pub(crate) default: Option<(Token![=], ParsedType)>,
}

impl ToTokens for ParsedSizeTypeParam {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let Self {
            attrs,
            options,
            ident,
            colon_token,
            bounds,
            default,
        } = self;
        let TypeParamOptions {} = options.body;
        for attr in attrs {
            attr.to_tokens(tokens);
        }
        ident.to_tokens(tokens);
        colon_token.to_tokens(tokens);
        bounds.to_tokens(tokens);
        if let Some((eq, ty)) = default {
            eq.to_tokens(tokens);
            ty.to_tokens(tokens);
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) enum ParsedGenericParam {
    Type(ParsedTypeParam),
    SizeType(ParsedSizeTypeParam),
}

impl ToTokens for ParsedGenericParam {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        match self {
            ParsedGenericParam::Type(v) => v.to_tokens(tokens),
            ParsedGenericParam::SizeType(v) => v.to_tokens(tokens),
        }
    }
}

impl ParsedGenericParam {
    pub(crate) fn ident(&self) -> &Ident {
        match self {
            ParsedGenericParam::Type(v) => &v.ident,
            ParsedGenericParam::SizeType(v) => &v.ident,
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedGenerics {
    pub(crate) lt_token: Option<Token![<]>,
    pub(crate) params: Punctuated<ParsedGenericParam, Token![,]>,
    pub(crate) gt_token: Option<Token![>]>,
    pub(crate) param_name_to_index_map: HashMap<Ident, usize>,
}

impl ParsedGenerics {
    pub(crate) fn for_static_type(mut self) -> ParsedGenerics {
        for param in self.params.iter_mut() {
            match param {
                ParsedGenericParam::Type(ParsedTypeParam { ident, bounds, .. }) => {
                    bounds
                        .StaticType
                        .get_or_insert_with(|| known_items::StaticType(ident.span()));
                }
                ParsedGenericParam::SizeType(ParsedSizeTypeParam { ident, bounds, .. }) => {
                    bounds
                        .KnownSize
                        .get_or_insert_with(|| known_items::KnownSize(ident.span()));
                }
            }
        }
        self
    }
    pub(crate) fn parse(generics: &mut Generics) -> syn::Result<Self> {
        let Generics {
            lt_token,
            params,
            gt_token,
            where_clause,
        } = generics;
        let mut errors = Errors::new();
        let mut predicates: Vec<WherePredicate> = Vec::with_capacity(params.len());
        let mut defaults = Vec::with_capacity(params.len());
        let mut params = Punctuated::<UnparsedGenericParam, _>::from_iter(
            params.pairs_mut().filter_map_pair_value_mut(|param| {
                let (param, default) = match param {
                    GenericParam::Lifetime(param) => {
                        errors.unwrap_or_default(
                            HdlAttr::<LifetimeParamOptions>::parse_and_take_attr(&mut param.attrs),
                        );
                        errors.error(param, "lifetime generics are not supported by #[hdl]");
                        return None;
                    }
                    GenericParam::Type(TypeParam {
                        attrs,
                        ident,
                        colon_token,
                        bounds,
                        eq_token,
                        default,
                    }) => {
                        let span = ident.span();
                        let options = errors
                            .unwrap_or_default(HdlAttr::<TypeParamOptions>::parse_and_take_attr(
                                attrs,
                            ))
                            .unwrap_or_default();
                        let colon_token = colon_token.unwrap_or_else(|| Token![:](span));
                        if !bounds.is_empty() {
                            predicates.push(WherePredicate::Type(PredicateType {
                                lifetimes: None,
                                bounded_ty: parse_quote! { #ident },
                                colon_token,
                                bounds: bounds.clone(),
                            }));
                        }
                        (
                            UnparsedGenericParam::Type {
                                attrs: attrs.clone(),
                                options,
                                ident: ident.clone(),
                                colon_token,
                                bounds: ParsedBounds::default(),
                            },
                            default
                                .clone()
                                .map(|v| (eq_token.unwrap_or_else(|| Token![=](span)), v)),
                        )
                    }
                    GenericParam::Const(ConstParam {
                        attrs,
                        const_token,
                        ident,
                        colon_token,
                        ty,
                        eq_token,
                        default,
                    }) => {
                        let options = errors
                            .unwrap_or_default(HdlAttr::<ConstParamOptions>::parse_and_take_attr(
                                attrs,
                            ))
                            .unwrap_or_default();
                        todo!()
                    }
                };
                defaults.push(default);
                Some(param)
            }),
        );
        let param_name_to_index_map: HashMap<Ident, usize> = params
            .iter()
            .enumerate()
            .map(|(index, param)| {
                let UnparsedGenericParam::Type { ident, .. } = param;
                (ident.clone(), index)
            })
            .collect();
        if let Some(where_clause) = where_clause {
            predicates.extend(where_clause.predicates.iter().cloned());
        }
        for predicate in predicates {
            let WherePredicate::Type(PredicateType {
                lifetimes: None,
                bounded_ty:
                    Type::Path(TypePath {
                        qself: None,
                        path: bounded_ty,
                    }),
                colon_token: _,
                bounds: unparsed_bounds,
            }) = predicate
            else {
                errors.error(predicate, "unsupported where predicate kind");
                continue;
            };
            let Some(&index) = bounded_ty
                .get_ident()
                .and_then(|bounded_ty| param_name_to_index_map.get(bounded_ty))
            else {
                errors.error(
                    bounded_ty,
                    "where predicate bounded type must be one of the generic parameters",
                );
                continue;
            };
            let UnparsedGenericParam::Type {
                bounds: parsed_bounds,
                ..
            } = &mut params[index];
            parsed_bounds.extend(errors.ok(syn::parse2::<ParsedBounds>(
                unparsed_bounds.to_token_stream(),
            )));
        }
        let params =
            Punctuated::from_iter(params.into_pairs().map_pair_value(|param| match param {
                UnparsedGenericParam::Type {
                    attrs,
                    options,
                    ident,
                    colon_token,
                    mut bounds,
                } => {
                    bounds.add_implied_bounds();
                    match bounds.categorize(&mut errors, ident.span()) {
                        ParsedBoundsCategory::Type(bounds) => {
                            ParsedGenericParam::Type(ParsedTypeParam {
                                attrs,
                                options,
                                ident,
                                colon_token,
                                bounds,
                                default: None,
                            })
                        }
                        ParsedBoundsCategory::SizeType(bounds) => {
                            ParsedGenericParam::SizeType(ParsedSizeTypeParam {
                                attrs,
                                options,
                                ident,
                                colon_token,
                                bounds,
                                default: None,
                            })
                        }
                    }
                }
            }));
        let mut retval = Self {
            lt_token: *lt_token,
            params,
            gt_token: *gt_token,
            param_name_to_index_map,
        };
        for (cur_param_index, default) in defaults.into_iter().enumerate() {
            let Some((eq, mut ty)) = default else {
                continue;
            };
            let Ok(ty) = TypesParser {
                generics: &retval,
                cur_param_index: Some(cur_param_index),
                errors: &mut errors,
            }
            .parse(&mut ty) else {
                continue;
            };
            match &mut retval.params[cur_param_index] {
                ParsedGenericParam::Type(ParsedTypeParam { default, .. })
                | ParsedGenericParam::SizeType(ParsedSizeTypeParam { default, .. }) => {
                    *default = Some((eq, ty))
                }
            }
        }
        errors.finish()?;
        Ok(retval)
    }
}

impl ToTokens for ParsedGenerics {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let Self {
            lt_token,
            params,
            gt_token,
            param_name_to_index_map: _,
        } = self;
        if params.is_empty() {
            return;
        }
        lt_token.unwrap_or_default().to_tokens(tokens);
        params.to_tokens(tokens);
        gt_token.unwrap_or_default().to_tokens(tokens);
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedGenericsImplGenerics<'a>(&'a ParsedGenerics);

impl ToTokens for ParsedGenericsImplGenerics<'_> {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let ParsedGenerics {
            lt_token,
            params,
            gt_token,
            param_name_to_index_map: _,
        } = &self.0;
        if params.is_empty() {
            return;
        }
        lt_token.unwrap_or_default().to_tokens(tokens);
        for param in params.pairs() {
            let (param, punct) = param.into_tuple();
            match param {
                ParsedGenericParam::Type(ParsedTypeParam {
                    attrs: _,
                    options,
                    ident,
                    colon_token,
                    bounds,
                    default: _,
                }) => {
                    let TypeParamOptions {} = options.body;
                    ident.to_tokens(tokens);
                    colon_token.to_tokens(tokens);
                    bounds.to_tokens(tokens);
                }
                ParsedGenericParam::SizeType(ParsedSizeTypeParam {
                    attrs: _,
                    options,
                    ident,
                    colon_token,
                    bounds,
                    default: _,
                }) => {
                    let TypeParamOptions {} = options.body;
                    ident.to_tokens(tokens);
                    colon_token.to_tokens(tokens);
                    bounds.to_tokens(tokens);
                }
            }
            punct.to_tokens(tokens);
        }
        gt_token.unwrap_or_default().to_tokens(tokens);
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedGenericsTurbofish<'a>(&'a ParsedGenerics);

impl ToTokens for ParsedGenericsTurbofish<'_> {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        if self.0.params.is_empty() {
            return;
        }
        Token![::](self.0.lt_token.unwrap_or_default().span).to_tokens(tokens);
        ParsedGenericsTypeGenerics(self.0).to_tokens(tokens);
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedGenericsTypeGenerics<'a>(&'a ParsedGenerics);

impl ToTokens for ParsedGenericsTypeGenerics<'_> {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let ParsedGenerics {
            lt_token,
            params,
            gt_token,
            param_name_to_index_map: _,
        } = &self.0;
        if params.is_empty() {
            return;
        }
        lt_token.unwrap_or_default().to_tokens(tokens);
        for param in params.pairs() {
            let (param, punct) = param.into_tuple();
            param.ident().to_tokens(tokens);
            punct.to_tokens(tokens);
        }
        gt_token.unwrap_or_default().to_tokens(tokens);
    }
}

#[derive(Debug, Clone)]
#[allow(dead_code)]
pub(crate) struct ParsedGenericsWhereClause<'a>(&'a ParsedGenerics);

impl ToTokens for ParsedGenericsWhereClause<'_> {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let _ = tokens;
    }
}

#[allow(dead_code)]
pub(crate) trait AsTurbofish {
    type Turbofish<'a>: 'a + ToTokens + Clone + fmt::Debug
    where
        Self: 'a;
    fn as_turbofish(&self) -> Self::Turbofish<'_>;
}

impl AsTurbofish for TypeGenerics<'_> {
    type Turbofish<'a> = Turbofish<'a> where Self: 'a;

    fn as_turbofish(&self) -> Self::Turbofish<'_> {
        TypeGenerics::as_turbofish(self)
    }
}

impl AsTurbofish for ParsedGenericsTypeGenerics<'_> {
    type Turbofish<'a> = ParsedGenericsTurbofish<'a>
    where
        Self: 'a;

    fn as_turbofish(&self) -> Self::Turbofish<'_> {
        ParsedGenericsTurbofish(self.0)
    }
}

pub(crate) trait SplitForImpl {
    type ImplGenerics<'a>: 'a + ToTokens + Clone + fmt::Debug
    where
        Self: 'a;
    type TypeGenerics<'a>: 'a + AsTurbofish + ToTokens + Clone + fmt::Debug
    where
        Self: 'a;
    type WhereClause<'a>: 'a + ToTokens + Clone + fmt::Debug
    where
        Self: 'a;
    fn split_for_impl(
        &self,
    ) -> (
        Self::ImplGenerics<'_>,
        Self::TypeGenerics<'_>,
        Self::WhereClause<'_>,
    );
}

impl SplitForImpl for Generics {
    type ImplGenerics<'a> = ImplGenerics<'a>;
    type TypeGenerics<'a> = TypeGenerics<'a>;
    type WhereClause<'a> = Option<&'a WhereClause>;
    fn split_for_impl(
        &self,
    ) -> (
        Self::ImplGenerics<'_>,
        Self::TypeGenerics<'_>,
        Self::WhereClause<'_>,
    ) {
        Generics::split_for_impl(&self)
    }
}

impl SplitForImpl for ParsedGenerics {
    type ImplGenerics<'a> = ParsedGenericsImplGenerics<'a>
    where
        Self: 'a;

    type TypeGenerics<'a> = ParsedGenericsTypeGenerics<'a>
    where
        Self: 'a;

    type WhereClause<'a> = ParsedGenericsWhereClause<'a>
    where
        Self: 'a;

    fn split_for_impl(
        &self,
    ) -> (
        Self::ImplGenerics<'_>,
        Self::TypeGenerics<'_>,
        Self::WhereClause<'_>,
    ) {
        (
            ParsedGenericsImplGenerics(self),
            ParsedGenericsTypeGenerics(self),
            ParsedGenericsWhereClause(self),
        )
    }
}

impl From<&'_ ParsedGenerics> for Generics {
    fn from(value: &ParsedGenerics) -> Self {
        let mut retval: Generics = parse_quote! { #value };
        let where_clause = ParsedGenericsWhereClause(value).to_token_stream();
        if !where_clause.is_empty() {
            retval.where_clause = Some(parse_quote! { #where_clause });
        }
        retval
    }
}

impl From<ParsedGenerics> for Generics {
    fn from(value: ParsedGenerics) -> Self {
        From::<&ParsedGenerics>::from(&value)
    }
}

#[derive(Debug, Clone)]
pub(crate) enum MaybeParsed<P, U> {
    Unrecognized(U),
    Parsed(P),
}

impl<P, U> MaybeParsed<P, U> {
    pub(crate) fn as_ref(&self) -> MaybeParsed<&P, &U> {
        match self {
            MaybeParsed::Unrecognized(v) => MaybeParsed::Unrecognized(v),
            MaybeParsed::Parsed(v) => MaybeParsed::Parsed(v),
        }
    }
    pub(crate) fn as_mut(&mut self) -> MaybeParsed<&mut P, &mut U> {
        match self {
            MaybeParsed::Unrecognized(v) => MaybeParsed::Unrecognized(v),
            MaybeParsed::Parsed(v) => MaybeParsed::Parsed(v),
        }
    }
    pub(crate) fn map<PF: FnOnce(P) -> PR, UF: FnOnce(U) -> UR, PR, UR>(
        self,
        pf: PF,
        uf: UF,
    ) -> MaybeParsed<PR, UR> {
        match self {
            MaybeParsed::Unrecognized(v) => MaybeParsed::Unrecognized(uf(v)),
            MaybeParsed::Parsed(v) => MaybeParsed::Parsed(pf(v)),
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) struct MaybeParsedIter<P, U>(pub(crate) MaybeParsed<P, U>);

impl<P: Iterator, U: Iterator> Iterator for MaybeParsedIter<P, U> {
    type Item = MaybeParsed<P::Item, U::Item>;

    fn next(&mut self) -> Option<Self::Item> {
        match &mut self.0 {
            MaybeParsed::Unrecognized(i) => i.next().map(MaybeParsed::Unrecognized),
            MaybeParsed::Parsed(i) => i.next().map(MaybeParsed::Parsed),
        }
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        match &self.0 {
            MaybeParsed::Unrecognized(i) => i.size_hint(),
            MaybeParsed::Parsed(i) => i.size_hint(),
        }
    }
}

impl<P: ExactSizeIterator, U: ExactSizeIterator> ExactSizeIterator for MaybeParsedIter<P, U> {}

impl<P: DoubleEndedIterator, U: DoubleEndedIterator> DoubleEndedIterator for MaybeParsedIter<P, U> {
    fn next_back(&mut self) -> Option<Self::Item> {
        match &mut self.0 {
            MaybeParsed::Unrecognized(i) => i.next_back().map(MaybeParsed::Unrecognized),
            MaybeParsed::Parsed(i) => i.next_back().map(MaybeParsed::Parsed),
        }
    }
}

impl<P: IntoIterator, U: IntoIterator> IntoIterator for MaybeParsed<P, U> {
    type Item = MaybeParsed<P::Item, U::Item>;
    type IntoIter = MaybeParsedIter<P::IntoIter, U::IntoIter>;

    fn into_iter(self) -> Self::IntoIter {
        MaybeParsedIter(match self {
            MaybeParsed::Unrecognized(v) => MaybeParsed::Unrecognized(v.into_iter()),
            MaybeParsed::Parsed(v) => MaybeParsed::Parsed(v.into_iter()),
        })
    }
}

impl<P, U> MaybeParsed<P, U> {
    pub(crate) fn iter<'a>(
        &'a self,
    ) -> MaybeParsedIter<<&'a P as IntoIterator>::IntoIter, <&'a U as IntoIterator>::IntoIter>
    where
        &'a P: IntoIterator,
        &'a U: IntoIterator,
    {
        self.into_iter()
    }
    pub(crate) fn iter_mut<'a>(
        &'a mut self,
    ) -> MaybeParsedIter<<&'a mut P as IntoIterator>::IntoIter, <&'a mut U as IntoIterator>::IntoIter>
    where
        &'a mut P: IntoIterator,
        &'a mut U: IntoIterator,
    {
        self.into_iter()
    }
}

impl<'a, P, U> IntoIterator for &'a MaybeParsed<P, U>
where
    &'a P: IntoIterator,
    &'a U: IntoIterator,
{
    type Item = MaybeParsed<<&'a P as IntoIterator>::Item, <&'a U as IntoIterator>::Item>;
    type IntoIter =
        MaybeParsedIter<<&'a P as IntoIterator>::IntoIter, <&'a U as IntoIterator>::IntoIter>;

    fn into_iter(self) -> Self::IntoIter {
        MaybeParsedIter(match self {
            MaybeParsed::Unrecognized(v) => MaybeParsed::Unrecognized(v.into_iter()),
            MaybeParsed::Parsed(v) => MaybeParsed::Parsed(v.into_iter()),
        })
    }
}

impl<'a, P, U> IntoIterator for &'a mut MaybeParsed<P, U>
where
    &'a mut P: IntoIterator,
    &'a mut U: IntoIterator,
{
    type Item = MaybeParsed<<&'a mut P as IntoIterator>::Item, <&'a mut U as IntoIterator>::Item>;
    type IntoIter = MaybeParsedIter<
        <&'a mut P as IntoIterator>::IntoIter,
        <&'a mut U as IntoIterator>::IntoIter,
    >;

    fn into_iter(self) -> Self::IntoIter {
        MaybeParsedIter(match self {
            MaybeParsed::Unrecognized(v) => MaybeParsed::Unrecognized(v.into_iter()),
            MaybeParsed::Parsed(v) => MaybeParsed::Parsed(v.into_iter()),
        })
    }
}

impl From<MaybeParsed<ParsedGenerics, Generics>> for Generics {
    fn from(value: MaybeParsed<ParsedGenerics, Generics>) -> Self {
        match value {
            MaybeParsed::Unrecognized(value) => value,
            MaybeParsed::Parsed(value) => value.into(),
        }
    }
}

impl From<&'_ MaybeParsed<ParsedGenerics, Generics>> for Generics {
    fn from(value: &MaybeParsed<ParsedGenerics, Generics>) -> Self {
        match value {
            MaybeParsed::Unrecognized(value) => value.clone(),
            MaybeParsed::Parsed(value) => value.into(),
        }
    }
}

impl<P: ToTokens, U: ToTokens> ToTokens for MaybeParsed<P, U> {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        match self {
            MaybeParsed::Unrecognized(v) => v.to_tokens(tokens),
            MaybeParsed::Parsed(v) => v.to_tokens(tokens),
        }
    }

    fn to_token_stream(&self) -> TokenStream {
        match self {
            MaybeParsed::Unrecognized(v) => v.to_token_stream(),
            MaybeParsed::Parsed(v) => v.to_token_stream(),
        }
    }

    fn into_token_stream(self) -> TokenStream {
        match self {
            MaybeParsed::Unrecognized(v) => v.into_token_stream(),
            MaybeParsed::Parsed(v) => v.into_token_stream(),
        }
    }
}

impl<P: AsTurbofish, U: AsTurbofish> AsTurbofish for MaybeParsed<P, U> {
    type Turbofish<'a> = MaybeParsed<P::Turbofish<'a>, U::Turbofish<'a>>
    where
        Self: 'a;

    fn as_turbofish(&self) -> Self::Turbofish<'_> {
        match self {
            MaybeParsed::Unrecognized(v) => MaybeParsed::Unrecognized(v.as_turbofish()),
            MaybeParsed::Parsed(v) => MaybeParsed::Parsed(v.as_turbofish()),
        }
    }
}

impl<P: SplitForImpl, U: SplitForImpl> SplitForImpl for MaybeParsed<P, U> {
    type ImplGenerics<'a> = MaybeParsed<P::ImplGenerics<'a>, U::ImplGenerics<'a>>
    where
        Self: 'a;
    type TypeGenerics<'a> = MaybeParsed<P::TypeGenerics<'a>, U::TypeGenerics<'a>>
    where
        Self: 'a;
    type WhereClause<'a> = MaybeParsed<P::WhereClause<'a>, U::WhereClause<'a>>
    where
        Self: 'a;

    fn split_for_impl(
        &self,
    ) -> (
        Self::ImplGenerics<'_>,
        Self::TypeGenerics<'_>,
        Self::WhereClause<'_>,
    ) {
        match self {
            MaybeParsed::Unrecognized(v) => {
                let (i, t, w) = v.split_for_impl();
                (
                    MaybeParsed::Unrecognized(i),
                    MaybeParsed::Unrecognized(t),
                    MaybeParsed::Unrecognized(w),
                )
            }
            MaybeParsed::Parsed(v) => {
                let (i, t, w) = v.split_for_impl();
                (
                    MaybeParsed::Parsed(i),
                    MaybeParsed::Parsed(t),
                    MaybeParsed::Parsed(w),
                )
            }
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedField {
    pub(crate) attrs: Vec<Attribute>,
    pub(crate) vis: Visibility,
    pub(crate) ident: Option<Ident>,
    pub(crate) colon_token: Option<Token![:]>,
    pub(crate) ty: ParsedType,
}

impl ParseTypes<Field> for ParsedField {
    fn parse_types(input: &mut Field, parser: &mut TypesParser<'_>) -> Result<Self, ParseFailed> {
        let Field {
            attrs,
            vis,
            mutability,
            ident,
            colon_token,
            ty,
        } = input;
        if !matches!(mutability, FieldMutability::None) {
            // FIXME: use mutability as the spanned tokens,
            // blocked on https://github.com/dtolnay/syn/issues/1717
            parser
                .errors()
                .error(&ident, "field mutability is not supported");
            *mutability = FieldMutability::None;
        }
        *mutability = FieldMutability::None;
        Ok(Self {
            attrs: attrs.clone(),
            vis: vis.clone(),
            ident: ident.clone(),
            colon_token: *colon_token,
            ty: parser.parse(ty)?,
        })
    }
}

impl From<ParsedField> for Field {
    fn from(value: ParsedField) -> Self {
        Self {
            attrs: value.attrs,
            vis: value.vis,
            mutability: FieldMutability::None,
            ident: value.ident,
            colon_token: value.colon_token,
            ty: value.ty.into(),
        }
    }
}

impl MaybeParsed<ParsedField, Field> {
    pub(crate) fn ident(&self) -> &Option<Ident> {
        match self {
            MaybeParsed::Unrecognized(v) => &v.ident,
            MaybeParsed::Parsed(v) => &v.ident,
        }
    }
}

impl<'a> MaybeParsed<&'a ParsedField, &'a Field> {
    pub(crate) fn ident(self) -> &'a Option<Ident> {
        match self {
            MaybeParsed::Unrecognized(v) => &v.ident,
            MaybeParsed::Parsed(v) => &v.ident,
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) struct ParsedFieldsNamed {
    pub(crate) brace_token: Brace,
    pub(crate) named: Punctuated<ParsedField, Token![,]>,
}

impl ParseTypes<FieldsNamed> for ParsedFieldsNamed {
    fn parse_types(
        input: &mut FieldsNamed,
        parser: &mut TypesParser<'_>,
    ) -> Result<Self, ParseFailed> {
        let FieldsNamed { brace_token, named } = input;
        Ok(Self {
            brace_token: *brace_token,
            named: parser.parse(named)?,
        })
    }
}

impl From<ParsedFieldsNamed> for FieldsNamed {
    fn from(value: ParsedFieldsNamed) -> Self {
        Self {
            brace_token: value.brace_token,
            named: value
                .named
                .into_pairs()
                .map_pair_value(Into::into)
                .collect(),
        }
    }
}

impl MaybeParsed<ParsedFieldsNamed, FieldsNamed> {
    pub(crate) fn named(
        &self,
    ) -> MaybeParsed<&Punctuated<ParsedField, Token![,]>, &Punctuated<Field, Token![,]>> {
        self.as_ref().map(|v| &v.named, |v| &v.named)
    }
}

impl From<MaybeParsed<ParsedFieldsNamed, FieldsNamed>> for FieldsNamed {
    fn from(value: MaybeParsed<ParsedFieldsNamed, FieldsNamed>) -> Self {
        match value {
            MaybeParsed::Unrecognized(value) => value,
            MaybeParsed::Parsed(value) => value.into(),
        }
    }
}
