// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    expr::{ops::VariantAccess, Expr, ToExpr},
    hdl,
    int::Bool,
    intern::{Intern, Interned},
    module::{
        EnumMatchVariantAndInactiveScopeImpl, EnumMatchVariantsIterImpl, ModuleBuilder,
        NormalModule, Scope,
    },
    source_location::SourceLocation,
    ty::{CanonicalType, MatchVariantAndInactiveScope, Type, TypeProperties},
};
use hashbrown::HashMap;
use std::iter::FusedIterator;

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct EnumVariant {
    pub name: Interned<str>,
    pub ty: Option<CanonicalType>,
}

#[derive(Clone, Eq)]
struct EnumImpl {
    variants: Interned<[EnumVariant]>,
    name_indexes: HashMap<Interned<str>, usize>,
    type_properties: TypeProperties,
}

impl std::fmt::Debug for EnumImpl {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Enum")
            .field("variants", &self.variants)
            .field("name_indexes", &self.name_indexes)
            .field("type_properties", &self.type_properties)
            .finish()
    }
}

impl std::hash::Hash for EnumImpl {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.variants.hash(state);
    }
}

impl PartialEq for EnumImpl {
    fn eq(&self, other: &Self) -> bool {
        self.variants == other.variants
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct Enum(Interned<EnumImpl>);

fn discriminant_bit_width_impl(variant_count: usize) -> usize {
    variant_count
        .next_power_of_two()
        .checked_ilog2()
        .unwrap_or(0) as usize
}

impl Enum {
    #[track_caller]
    pub fn new(variants: Interned<[EnumVariant]>) -> Self {
        assert!(!variants.is_empty(), "zero-variant enums aren't yet supported: https://github.com/chipsalliance/firrtl-spec/issues/208");
        let mut name_indexes = HashMap::with_capacity(variants.len());
        let mut type_properties = TypeProperties {
            is_passive: true,
            is_storable: true,
            is_castable_from_bits: true,
            bit_width: 0,
        };
        for (index, EnumVariant { name, ty }) in variants.iter().enumerate() {
            if let Some(old_index) = name_indexes.insert(*name, index) {
                panic!("duplicate variant name {name:?}: at both index {old_index} and {index}");
            }
            if let Some(ty) = ty {
                let TypeProperties {
                    is_passive,
                    is_storable,
                    is_castable_from_bits,
                    bit_width,
                } = ty.type_properties();
                assert!(is_passive, "variant type must be a passive type: {ty:?}");
                type_properties = TypeProperties {
                    is_passive: true,
                    is_storable: type_properties.is_storable & is_storable,
                    is_castable_from_bits: type_properties.is_castable_from_bits
                        & is_castable_from_bits,
                    bit_width: type_properties.bit_width.max(bit_width),
                };
            }
        }
        type_properties.bit_width = type_properties
            .bit_width
            .checked_add(discriminant_bit_width_impl(variants.len()))
            .unwrap_or_else(|| panic!("enum is too big: bit-width overflowed"));
        Self(
            EnumImpl {
                variants,
                name_indexes,
                type_properties,
            }
            .intern_sized(),
        )
    }
    pub fn discriminant_bit_width(self) -> usize {
        discriminant_bit_width_impl(self.variants().len())
    }
    pub fn type_properties(self) -> TypeProperties {
        self.0.type_properties
    }
    pub fn name_indexes(&self) -> &HashMap<Interned<str>, usize> {
        &self.0.name_indexes
    }
}

pub trait EnumType:
    Type<
    MaskType = Bool,
    MatchActiveScope = Scope,
    MatchVariantAndInactiveScope = EnumMatchVariantAndInactiveScope<Self>,
    MatchVariantsIter = EnumMatchVariantsIter<Self>,
>
{
    fn variants(&self) -> Interned<[EnumVariant]>;
    fn match_activate_scope(
        v: Self::MatchVariantAndInactiveScope,
    ) -> (Self::MatchVariant, Self::MatchActiveScope);
}

pub struct EnumMatchVariantAndInactiveScope<T: EnumType>(EnumMatchVariantAndInactiveScopeImpl<T>);

impl<T: EnumType> MatchVariantAndInactiveScope for EnumMatchVariantAndInactiveScope<T> {
    type MatchVariant = T::MatchVariant;
    type MatchActiveScope = Scope;

    fn match_activate_scope(self) -> (Self::MatchVariant, Self::MatchActiveScope) {
        T::match_activate_scope(self)
    }
}

impl<T: EnumType> EnumMatchVariantAndInactiveScope<T> {
    pub fn variant_access(&self) -> Interned<VariantAccess> {
        self.0.variant_access()
    }
    pub fn activate(self) -> (Interned<VariantAccess>, Scope) {
        self.0.activate()
    }
}

#[derive(Clone)]
pub struct EnumMatchVariantsIter<T: EnumType> {
    pub(crate) inner: EnumMatchVariantsIterImpl<T>,
    pub(crate) variant_index: std::ops::Range<usize>,
}

impl<T: EnumType> Iterator for EnumMatchVariantsIter<T> {
    type Item = EnumMatchVariantAndInactiveScope<T>;

    fn next(&mut self) -> Option<Self::Item> {
        self.variant_index.next().map(|variant_index| {
            EnumMatchVariantAndInactiveScope(self.inner.for_variant_index(variant_index))
        })
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        self.variant_index.size_hint()
    }
}

impl<T: EnumType> ExactSizeIterator for EnumMatchVariantsIter<T> {
    fn len(&self) -> usize {
        self.variant_index.len()
    }
}

impl<T: EnumType> FusedIterator for EnumMatchVariantsIter<T> {}

impl<T: EnumType> DoubleEndedIterator for EnumMatchVariantsIter<T> {
    fn next_back(&mut self) -> Option<Self::Item> {
        self.variant_index.next_back().map(|variant_index| {
            EnumMatchVariantAndInactiveScope(self.inner.for_variant_index(variant_index))
        })
    }
}

impl EnumType for Enum {
    fn match_activate_scope(
        v: Self::MatchVariantAndInactiveScope,
    ) -> (Self::MatchVariant, Self::MatchActiveScope) {
        let (expr, scope) = v.0.activate();
        (expr.variant_type().map(|_| expr.to_expr()), scope)
    }
    fn variants(&self) -> Interned<[EnumVariant]> {
        self.0.variants
    }
}

impl Type for Enum {
    type MaskType = Bool;
    type MatchVariant = Option<Expr<CanonicalType>>;
    type MatchActiveScope = Scope;
    type MatchVariantAndInactiveScope = EnumMatchVariantAndInactiveScope<Self>;
    type MatchVariantsIter = EnumMatchVariantsIter<Self>;

    fn match_variants(
        this: Expr<Self>,
        module_builder: &mut ModuleBuilder<NormalModule>,
        source_location: crate::source_location::SourceLocation,
    ) -> Self::MatchVariantsIter {
        module_builder.enum_match_variants_helper(this, source_location)
    }

    fn mask_type(&self) -> Self::MaskType {
        Bool
    }

    fn canonical(&self) -> CanonicalType {
        CanonicalType::Enum(*self)
    }

    #[track_caller]
    fn from_canonical(canonical_type: CanonicalType) -> Self {
        let CanonicalType::Enum(retval) = canonical_type else {
            panic!("expected enum");
        };
        retval
    }
    fn source_location() -> SourceLocation {
        SourceLocation::builtin()
    }
}

#[hdl]
pub enum HdlOption<T: Type> {
    None,
    Some(T),
}
